@@ -2,7 +2,7 @@ import diffrax
 import jax.numpy as jnp
 import pytest
 
-from helpers import shaped_allclose
+from .helpers import shaped_allclose
 
 
 def _test(ts, flip=False):
